import dolfin
import numpy as np

try:
    from dolfin_adjoint import (
        Constant,
        Function,
        FunctionAssigner,
        assemble,
        interpolate,
        project,
    )
except ImportError:
    from dolfin import (
        Function,
        interpolate,
        Constant,
        project,
        assemble,
        FunctionAssigner,
    )

from . import numpy_mpi, utils
from .utils import DOLFIN_VERSION_MAJOR, logger, mpi_comm_world


def map_vector_field(f0, new_mesh, u=None, name="fiber", normalize=True):
    """
    Map a vector field (f0) onto a new mesh (new_mesh) where the new mesh
    can be a moved version of the original one according to some
    displacement (u). In that case we will just a Piola transform to
    map the vector field.
    """

    if DOLFIN_VERSION_MAJOR > 2016:
        dolfin.parameters["form_compiler"]["representation"] = "quadrature"

    dolfin.parameters["form_compiler"]["quadrature_degree"] = 4

    ufl_elem = f0.function_space().ufl_element()
    f0_new = dolfin.Function(dolfin.FunctionSpace(new_mesh, ufl_elem))

    if u is not None:

        f0_mesh = f0.function_space().mesh()
        u_elm = u.function_space().ufl_element()
        V = dolfin.FunctionSpace(f0_mesh, u_elm)
        u0 = dolfin.Function(V)
        # arr = numpy_mpi.gather_vector(u.vector())
        # numpy_mpi.assign_to_vector(u0.vector(), arr)
        u0.vector()[:] = u.vector()
        from .kinematics import DeformationGradient

        F = DeformationGradient(u0)

        f0_updated = dolfin.project(F * f0, f0.function_space())

        if normalize:
            f0_updated = normalize_vector_field(f0_updated)

        f0_new.vector()[:] = f0_updated.vector()
        # f0_arr = numpy_mpi.gather_vector(f0_updated.vector())
        # numpy_mpi.assign_to_vector(f0_new.vector(), f0_arr)

    else:
        # f0_arr = numpy_mpi.gather_vector(f0.vector())
        # numpy_mpi.assign_to_vector(f0_new.vector(), f0_arr)
        f0_new.vector()[:] = f0.vector()

    if DOLFIN_VERSION_MAJOR > 2016:
        dolfin.parameters["form_compiler"]["representation"] = "uflacs"

    return f0_new


def update_function(mesh, f):
    """Given a function :math:`f` defined on some domain,
    update the function so that it now is defined on the domain
    given in the mesh
    """

    f_new = Function(dolfin.FunctionSpace(mesh, f.ufl_element()))
    numpy_mpi.assign_to_vector(f_new.vector(), numpy_mpi.gather_vector(f.vector()))
    return f_new


def normalize_vector_field(u):
    """Given a vector field, return a vector field with an L2 norm equal to 1.0
    """
    dim = len(u)
    S = u.function_space().sub(0).collapse()

    components = vectorfield_to_components(u, S, dim)

    normarray = np.sqrt(sum(components[i].vector().norm("l2") ** 2 for i in range(dim)))

    for comp in components:
        comp.vector()[:] = comp.vector() / normarray

    assigners = [FunctionAssigner(u.function_space().sub(i), S) for i in range(dim)]
    for i, comp, assigner in zip(range(dim), components, assigners):
        assigner.assign(u.sub(i), comp)

    return u


def vectorfield_to_components(u, S, dim):
    components = [dolfin.Function(S) for i in range(dim)]
    assigners = [FunctionAssigner(S, u.function_space().sub(i)) for i in range(dim)]
    for i, comp, assigner in zip(range(dim), components, assigners):
        assigner.assign(comp, u.sub(i))

    return components


def get_pressure(problem):
    """Returns p_lv (and p_rv if BiV mesh)
    """

    plv = [p.traction for p in problem.bcs.neumann if p.name == "lv"]
    prv = [p.traction for p in problem.bcs.neumann if p.name == "rv"]

    assert len(plv) > 0, "Problem has no Neumann BC for LV endo"
    pressure = [plv[0]]
    if prv:
        pressure.append(prv[0])
        return tuple(pressure)
    else:
        return pressure[0]


def read_hdf5(h5name, func, h5group="", comm=mpi_comm_world()):

    try:
        with dolfin.HDF5File(comm, h5name, "r") as h5file:

            h5file.read(func, h5group)

    except IOError as ex:
        logger.error(ex)
        logger.error("Make sure file {} exist".format(h5name))
        raise ex

    except RuntimeError as ex:
        logger.errro(ex)
        logger.error(
            (
                "Something went wrong when reading file "
                "{h5name} into function {func} from group "
                "{h5group}"
            ).format(h5name=h5name, func=func, h5group=h5group)
        )
        raise ex


def map_displacement(u, old_space, new_space, approx, name="mapped displacement"):

    if approx == "interpolate":
        # Do we need dolfin-adjoint here or is dolfin enough?
        u_int = interpolate(project(u, old_space), new_space)  # , name=name)

    elif approx == "project":
        # Do we need dolfin-adjoint here or is dolfin enough?
        u_int = project(u, new_space)  # , name=name)

    else:
        u_int = u

    return u_int


def compute_meshvolume(domain=None, dx=dolfin.dx, subdomain_id=None):
    return Constant(
        dolfin.assemble(
            dolfin.Constant(1.0) * dx(domain=domain, subdomain_id=subdomain_id)
        )
    )


def get_cavity_volume(geometry, unload=False, chamber="lv", u=None, xshift=0.0):

    if unload:
        mesh = geometry.original_geometry
        ffun = dolfin.MeshFunction("size_t", mesh, 2, mesh.domains())
    else:
        mesh = geometry.mesh
        ffun = geometry.ffun

    if chamber == "lv":
        if "ENDO_LV" in geometry.markers:
            endo_marker = geometry.markers["ENDO_LV"]
        else:
            endo_marker = geometry.markers["ENDO"]

    else:
        endo_marker = geometry.markers["ENDO_RV"]

    if hasattr(endo_marker, "__len__"):
        endo_marker = endo_marker[0]

    ds = dolfin.Measure("exterior_facet", subdomain_data=ffun, domain=mesh)(endo_marker)

    vol_form = get_cavity_volume_form(geometry.mesh, u, xshift)
    return assemble(vol_form * ds)


def get_cavity_volume_form(mesh, u=None, xshift=0.0):

    from . import kinematics

    shift = Constant((xshift, 0.0, 0.0))
    X = dolfin.SpatialCoordinate(mesh) - shift
    N = dolfin.FacetNormal(mesh)

    if u is None:
        vol_form = (-1.0 / 3.0) * dolfin.dot(X, N)
    else:
        F = kinematics.DeformationGradient(u)
        J = kinematics.Jacobian(F)
        vol_form = (-1.0 / 3.0) * dolfin.dot(X + u, J * dolfin.inv(F).T * N)

    return vol_form


def get_constant(val, value_size=None, value_rank=0, constant=Constant):

    if isinstance(val, (Constant, dolfin.Constant)):
        return val
    elif isinstance(val, (Function, dolfin.Function)):
        arr = numpy_mpi.gather_vector(val.vector())
        return constant(arr)
    elif isinstance(val, dolfin.GenericVector):
        arr = numpy_mpi.gather_vector(val)
        return constant(arr)

    if value_size is None:
        if np.isscalar(val):
            value_size = 1
        else:
            try:
                value_size = len(val)
                val = np.array(val)
            except Exception as ex:
                logger.debug(ex)
                # Hope for the best
                value_size = 1
    if value_size == 1:
        if value_rank == 0:
            c = constant(val)
        else:
            c = constant([val])
    else:
        c = constant([val] * value_size)

    return c


def get_dimesion(u):

    # TODO : Check argument
    try:
        if DOLFIN_VERSION_MAJOR > 1.6:
            from ufl.domain import find_geometric_dimension

            dim = find_geometric_dimension(u)
        else:
            dim = u.geometric_dimension()

    except Exception as ex:

        try:
            dim = len(u)
        except Exception as ex2:
            logger.warning(ex)
            logger.warning(ex2)
            # Assume dimension is 3
            logger.warning("Assume dimension is 3")
            dim = 3

    return dim


def subplus(x):
    r"""
    Ramp function

    .. math::

       \max\{x,0\}

    """

    return dolfin.conditional(dolfin.ge(x, 0.0), x, 0.0)


def heaviside(x):
    r"""
    Heaviside function

    .. math::

       \frac{\mathrm{d}}{\mathrm{d}x} \max\{x,0\}

    """

    return dolfin.conditional(dolfin.ge(x, 0.0), 1.0, 0.0)


def list_sum(lst):
    """
    Return the sum of a list, when the convetiional
    method (like `sum`) it not working.
    For example if you have a list of dolfin functions.

    :param list l: a list of objects
    :returns: The sum of the list. The type depends on
              the type of elemets in the list

    """

    if not isinstance(lst, list):
        return lst

    out = lst[0]
    for item in lst[1:]:
        out += item
    return out


def get_spaces(mesh):
    """
    Return an object of dolfin FunctionSpace, to
    be used in the optimization pipeline

    :param mesh: The mesh
    :type mesh: :py:class:`dolfin.Mesh`
    :returns: An object of functionspaces
    :rtype: object

    """

    # Make a dummy object
    spaces = utils.Object()

    # A real space with scalars used for dolfin adjoint
    spaces.r_space = dolfin.FunctionSpace(mesh, "R", 0)

    # A space for the strain fields
    spaces.strainfieldspace = dolfin.VectorFunctionSpace(mesh, "CG", 1, dim=3)

    # A space used for scalar strains
    spaces.strainspace = dolfin.VectorFunctionSpace(mesh, "R", 0, dim=3)

    # Spaces for the strain weights
    spaces.strain_weight_space = dolfin.TensorFunctionSpace(mesh, "R", 0)

    return spaces


def QuadratureSpace(mesh, degree, dim=3):
    """
    From FEniCS version 1.6 to 2016.1 there was a change in how
    FunctionSpace is defined for quadrature spaces.
    This functions checks your dolfin version and returns the correct
    quadrature space

    :param mesh: The mesh
    :type mesh: :py:class:`dolfin.Mesh`
    :param int degree: The degree of the element
    :param int dim: For a mesh of topological dimension 3,
                    dim = 1 would be a scalar function, and
                    dim = 3 would be a vector function.
    :returns: The quadrature space
    :rtype: :py:class:`dolfin.FunctionSpace`

    """

    if DOLFIN_VERSION_MAJOR > 1.6:
        if dim == 1:
            element = dolfin.FiniteElement(
                family="Quadrature",
                cell=mesh.ufl_cell(),
                degree=degree,
                quad_scheme="default",
            )
        else:
            element = dolfin.VectorElement(
                family="Quadrature",
                cell=mesh.ufl_cell(),
                degree=degree,
                quad_scheme="default",
            )

        return dolfin.FunctionSpace(mesh, element)
    else:
        if dim == 1:
            return dolfin.FunctionSpace(mesh, "Quadrature", degree)
        else:
            return dolfin.VectorFunctionSpace(mesh, "Quadrature", degree)


class VertexDomain(dolfin.SubDomain):
    """
    A subdomain defined in terms of
    a given set of coordinates.
    A point that is close to the given coordinates
    within a given tolerance will be marked as inside
    the domain.
    """

    def __init__(self, coords, tol=1e-4):
        """
        *Arguments*
          coords (list)
            List of coordinates for vertices in reference geometry
            defining this domains

          tol (float)
            Tolerance for how close a pointa should be to the given coordinates
            to be marked as inside the domain
        """

        self.coords = np.array(coords)
        self.tol = tol
        dolfin.SubDomain.__init__(self)

    def inside(self, x, on_boundary):

        if np.all([np.any(abs(x[i] - self.coords.T[i]) < self.tol) for i in range(3)]):
            return True

        return False


class BaseExpression(dolfin.Expression):
    """
    A class for assigning boundary condition according to segmented surfaces
    Since the base is located at x = a (usually a=0), two classes must be set:
    One for the y-direction and one for the z-direction

    Point on the endocardium and epicardium is given and the
    points on the mesh base is set accordingly.
    Points that lie on the base but not on the epi- or endoring
    will be given a zero value.
    """

    def __init__(self, mesh_verts, seg_verts, sub, it, name):
        """

        *Arguments*
          mesh: (dolfin.mesh)
            The mesh

          u: (dolfin.GenericFunction)
            Initial displacement

          mesh_verts (numpy.ndarray or list)
            Point of endocardial base from mesh

          seg_verts (numpy.ndarray or list)
            Point of endocardial base from segmentation

          sub (str)
            Either "y" or "z". The displacement in this direction is returned

          it (dolfin.Expression)
            Can be used to incrment the direclet bc

        """

        assert sub in ["y", "z"]
        self._mesh_verts = np.array(mesh_verts)
        self._all_seg_verts = np.array(seg_verts)
        self.point = 0
        self.npoints = len(seg_verts) - 1

        self._seg_verts = self._all_seg_verts[0]

        self._sub = sub
        self._it = it
        self.rename(name, name)

    def next(self):
        self._it.t = 0
        self.point += 1
        self._seg_verts = self._all_seg_verts[self.point]

    def reset(self):
        self.point = 0
        self._it.t = 0

    def eval(self, value, x):

        # Check if given coordinate is in the endoring vertices
        # and find the cooresponding index
        d = [np.where(x[i] == self._mesh_verts.T[i])[0] for i in range(3)]
        d_intersect = set.intersection(*map(set, d))
        assert len(d_intersect) < 2
        if len(d_intersect) == 1:

            idx = d_intersect.pop()
            prev_seg_verts = self._all_seg_verts[self.point - 1]

            # Return the displacement in the given direction
            # Iterated starting from the previous displacemet
            # to the current one
            if self._sub == "y":
                u_prev = self._mesh_verts[idx][1] - prev_seg_verts[idx][1]
                u_current = self._mesh_verts[idx][1] - self._seg_verts[idx][1]
                # value[0] = u_prev + self._it.t*(u_current - u_prev)
            else:  # sub == "z"
                u_prev = self._mesh_verts[idx][2] - prev_seg_verts[idx][2]
                u_current = self._mesh_verts[idx][2] - self._seg_verts[idx][2]

            val = u_prev + self._it.t * (u_current - u_prev)
            value[0] = val

        else:
            value[0] = 0


class MixedParameter(dolfin.Function):
    def __init__(self, fun, n, name="material_parameters"):
        """
        Initialize Mixed parameter.

        This will instanciate a function in a dolfin.MixedFunctionSpace
        consiting of `n` subspaces of the same type as `fun`.
        This is of course easy for the case when `fun` is a normal
        dolfin function, but in the case of a `RegionalParameter` it
        is not that straight forward.
        This class handles this case as well.


        :param fun: The type of you want to make a du
        :type fun: (:py:class:`dolfin.Function`)
        :param int n: number of subspaces
        :param str name: Name of the function

        .. todo::

           Implement support for MixedParameter with different
           types of subspaces, e.g [RegionalParamter, R_0, CG_1]

        """

        msg = "Please provide a dolin function as argument to MixedParameter"
        assert isinstance(fun, (dolfin.Function, Function, RegionalParameter)), msg

        if isinstance(fun, RegionalParameter):
            raise NotImplementedError

        # We can just make a usual mixed function space
        # with n copies of the original one
        V = fun.function_space()
        W = dolfin.MixedFunctionSpace([V] * n)

        Function.__init__(self, W, name=name)

        # Create a function assigner
        self.function_assigner = [FunctionAssigner(W.sub(i), V) for i in range(n)]

        # Store the original function space
        self.basespace = V

        if isinstance(fun, RegionalParameter):
            self._meshfunction = fun._meshfunction

    def assign_sub(self, f, i):
        """
        Assign subfunction

        :param f: The function you want to assign
        :param int i: The subspace number

        """
        f_ = dolfin.Function(self.basespace)
        f_.assign(f)
        self.function_assigner[i].assign(self.split()[i], f_)


class RegionalParameter(dolfin.Function):
    """A regional paramerter defined in terms of a dolfin.MeshFunction

    Suppose you have a MeshFunction defined different regions in your mesh,
    and you want to define different parameters on different regions,
    then this is what you want.
    """

    def __init__(self, meshfunction):

        # assert isinstance(
        #     meshfunction, dolfin.MeshFunctionSizet
        # ), "Invalid meshfunction for regional gamma"

        mesh = meshfunction.mesh()

        # FIXME
        self._values = set(numpy_mpi.gather_broadcast(meshfunction.array()))
        self._nvalues = len(self._values)

        V = dolfin.VectorFunctionSpace(mesh, "R", 0, dim=self._nvalues)

        dolfin.Function.__init__(self, V)
        self._meshfunction = meshfunction

        # Functionspace for the indicator functions
        self._proj_space = dolfin.FunctionSpace(mesh, "DG", 0)

        # Make indicator functions
        self._ind_functions = []
        for v in self._values:
            self._ind_functions.append(self._make_indicator_function(v))

    def get_values(self):
        return self._values

    @property
    def function(self):
        """
        Return linear combination of coefficents
        and basis functions

        :returns: A function with parameter values at each segment
                  specified by the meshfunction
        :rtype:  :py:class`dolfin.Function

        """
        return self._sum()

    @property
    def proj_space(self):
        """
        Space for projecting the scalars.
        This is a DG 0 space.
        """
        return self._proj_space

    def _make_indicator_function(self, marker):

        dofs = self._meshfunction.where_equal(marker)
        f = dolfin.Function(self._proj_space)
        f.vector()[dofs] = 1.0
        return f

    def _sum(self):
        coeffs = dolfin.split(self)
        fun = coeffs[0] * self._ind_functions[0]

        for c, f in zip(coeffs[1:], self._ind_functions[1:]):
            fun += c * f

        return fun



import dolfin
from .. import kinematics
from ..dolfin_utils import get_dimesion
from .active_model import ActiveModel, check_component


class ActiveStrain(ActiveModel):
    """
    This class implements the elastic invariants within
    the active strain framework

    Assuming transversally isotropic material for now

    """

    _model = "active_strain"

    @property
    def _mgamma(self):
        gamma = self.activation_field

        # FIXME: should allow for different active strain models
        if 1:
            mgamma = 1 - gamma
        elif self._model == "rossi":
            mgamma = 1 + gamma

        return mgamma

    def I1(self, F):

        I1 = self._I1(F)
        f0 = self.f0
        I4f = self._I4(F, f0)

        d = get_dimesion(F)
        mgamma = self._mgamma

        I1e = pow(mgamma, 4 - d) * I1 + (1 / mgamma ** 2 - pow(mgamma, 4 - d)) * I4f

        return I1e

    def I4(self, F, component="f0"):
        r"""
        Quasi-invariant in the elastic configuration
        Let :math:`d` be the geometric dimension.
        If

        .. math::

           \mathbf{F}_a = (1 - \gamma) \mathbf{f}_0 \otimes \mathbf{f}_0  +
           \frac{1}{\sqrt{1 - \gamma}}
           (\mathbf{I} - \mathbf{f}_0 \otimes \mathbf{f}_0)

        then

        .. math::

           I_{4f_0}^E = I_{4f_0} \frac{1}{(1+\gamma)^2}

        If

        .. math::

           \mathbf{F}_a = (1 + \gamma) \mathbf{f}_0 \otimes \mathbf{f}_0  +
           \frac{1}{\sqrt{1 + \gamma}}
           (\mathbf{I} - \mathbf{f}_0 \otimes \mathbf{f}_0)

        then

        .. math::

           I_{4f_0}^E = I_{4f_0} \frac{1}{(1+\gamma)^2}


        """

        check_component(component)
        a0 = getattr(self, component)
        I4f = self._I4(F, a0)
        mgamma = self._mgamma

        I4a0 = 1 / mgamma ** 2 * I4f

        return I4a0

    @property
    def Fa(self):

        f0 = self.f0
        d = get_dimesion(f0)
        f0f0 = dolfin.outer(f0, f0)
        I = kinematics.SecondOrderIdentity(f0f0)

        mgamma = self._mgamma
        Fa = mgamma * f0f0 + pow(mgamma, -1.0 / float(d - 1)) * (I - f0f0)

        return Fa

    def Fe(self, F):

        Fa = self.Fa
        Fe = F * dolfin.inv(Fa)

        return Fe



from .material_model import Material


class NeoHookean(Material):
    """
    Class for Neo Hookean material
    """

    name = "neo_hookean"

    @staticmethod
    def default_parameters():
        return {"mu": 15.0}

    def W_1(self, I_1, diff=0, dim=3, *args, **kwargs):

        mu = self.mu

        if diff == 0:
            return 0.5 * mu * (I_1 - dim)
        elif diff == 1:
            return 0.5 * mu
        elif diff == 2:
            return 0



#!/usr/bin/env python
# c) 2001-2017 Simula Research Laboratory ALL RIGHTS RESERVED
# Authors: Henrik Finsberg
# END-USER LICENSE AGREEMENT
# PLEASE READ THIS DOCUMENT CAREFULLY. By installing or using this
# software you agree with the terms and conditions of this license
# agreement. If you do not accept the terms of this license agreement
# you may not install or use this software.

# Permission to use, copy, modify and distribute any part of this
# software for non-profit educational and research purposes, without
# fee, and without a written agreement is hereby granted, provided
# that the above copyright notice, and this license agreement in its
# entirety appear in all copies. Those desiring to use this software
# for commercial purposes should contact Simula Research Laboratory AS:
# post@simula.no
#
# IN NO EVENT SHALL SIMULA RESEARCH LABORATORY BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE
# "PULSE-ADJOINT" EVEN IF SIMULA RESEARCH LABORATORY HAS BEEN ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE. THE SOFTWARE PROVIDED HEREIN IS
# ON AN "AS IS" BASIS, AND SIMULA RESEARCH LABORATORY HAS NO OBLIGATION
# TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# SIMULA RESEARCH LABORATORY MAKES NO REPRESENTATIONS AND EXTENDS NO
# WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESSED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
import dolfin

try:
    from dolfin_adjoint import Constant, Function, project
except ImportError:
    from dolfin import Constant, Function, project

from .. import kinematics
from .. import numpy_mpi
from ..dolfin_utils import get_dimesion, RegionalParameter, update_function

from .active_strain import ActiveStrain
from .active_stress import ActiveStress


def compressibility(model, *args, **kwargs):

    if model == "incompressible":
        return incompressible(*args, **kwargs)


def incompressible(p, J):
    return -p * (J - 1.0)


class Material(object):
    """
    Initialize material model

    Parameters
    ----------

    f0 : :py:class`dolfin.Function`
        Fiber field
    gamma : :py:class`dolfin.Function`
        Activation parameter
    params : dict
        Material parameters
    active_model : str
        Active model - active strain or active stress
    s0 : :py:class`dolfin.Function`
        Sheets
    n0 : :py:class`dolfin.Function`
        Sheet - normal
    T_ref : float
        Scale factor for activation parameter (default = 1.0)
    dev_iso_split : bool
        Decouple deformation into deviatoric and isochoric deformations
    eta : float
        Fraction of transverse active tesion for active stress formulation.
        0 = active only along fiber, 1 = equal forces in all directions
        (default=0.0).
    """

    def __init__(
        self,
        activation=None,
        parameters=None,
        active_model="active_strain",
        T_ref=None,
        eta=0.0,
        isochoric=True,
        compressible_model="incompressible",
        geometry=None,
        f0=None,
        s0=None,
        n0=None,
        *args,
        **kwargs
    ):

        # Parameters
        if parameters is None:
            parameters = self.default_parameters()

        self.parameters = parameters
        self._set_parameter_attrs(geometry)

        # Active model
        assert active_model in [
            "active_stress",
            "active_strain",
        ], "The active model '{}' is not implemented.".format(active_model)

        active_args = (activation, f0, s0, n0, T_ref, isochoric)
        # Activation
        if active_model == "active_stress":
            self.active = ActiveStress(*active_args, eta=eta, **kwargs)
        else:
            self.active = ActiveStrain(*active_args)

        self.compressible_model = compressible_model

        try:
            activation_element = self.activation.ufl_element()
            cell = activation_element.cell()
        except AttributeError:
            if not hasattr(activation, "cell"):
                cell = None
            else:
                cell = activation.cell()

        if geometry is not None and cell is not None:
            self.activation = update_function(geometry.mesh, self.activation)

    def _set_parameter_attrs(self, geometry=None):
        for k, v in self.parameters.items():

            if isinstance(v, (float, int)):
                setattr(self, k, Constant(v, name=k))

            elif isinstance(v, RegionalParameter):

                if geometry is not None:

                    v_new = RegionalParameter(geometry.sfun)
                    numpy_mpi.assign_to_vector(
                        v_new.vector(), numpy_mpi.gather_vector(v.vector())
                    )
                    v = v_new

                ind_space = v.proj_space
                setattr(self, k, Function(ind_space, name=k))
                mat = getattr(self, k)
                matfun = v.function
                mat.assign(dolfin.project(matfun, ind_space))

            else:

                if geometry is not None and v.ufl_element().cell() is not None:
                    v_new = update_function(geometry.mesh, v)
                    v = v_new

                setattr(self, k, v)

    def update_geometry(self, geometry):

        # Loop over the possible attributes containing
        # a domain and update it
        for m in ("f0", "s0", "n0"):
            setattr(self, m, getattr(geometry, m))

        self._set_parameter_attrs(geometry)

        activation_element = self.activation.ufl_element()
        if activation_element.cell() is not None:

            self.activation = update_function(geometry.mesh, self.activation)

    def copy(self, geometry=None):

        if geometry is not None:
            f0, s0, n0 = geometry.f0, geometry.s0, geometry.n0
        else:
            f0, s0, n0 = self.f0, self.s0, self.n0

        return self.__class__(
            activation=self.activation,
            parameters=self.parameters,
            active_model=self.active_model,
            geometry=geometry,
            f0=f0,
            s0=s0,
            n0=n0,
            T_ref=float(self.active.T_ref),
        )

    @property
    def name(self):
        return "generic_material"

    @property
    def f0(self):
        return self.active.f0

    @f0.setter
    def f0(self, f0):
        self.active.f0 = f0

    @property
    def s0(self):
        return self.active.s0

    @s0.setter
    def s0(self, s0):
        self.active.s0 = s0

    @property
    def n0(self):
        return self.active.n0

    @n0.setter
    def n0(self, n0):
        self.active.n0 = n0

    def __repr__(self):
        return (
            "{self.__class__.__name__}({self.parameters}, "
            "{self.active._model}, "
            "{self.compressible_model})"
        ).format(self=self)

    def compressibility(self, p, J):

        return compressibility(self.compressible_model, p, J)

    @property
    def active_model(self):
        return self.active.model_type

    @property
    def material_model(self):
        return self._model

    @property
    def is_isochoric(self):
        return self.active.is_isochoric

    @property
    def activation(self):
        """
        Return the activation paramter.
        If regional, this will return one parameter
        for each segment.
        """
        return self.active.activation

    @activation.setter
    def activation(self, f):
        self.active.activation = f

    @property
    def activation_field(self):
        """
        Return the contraciton paramter.
        If regional, this will return a piecewise
        constant function (DG_0)
        """
        return self.active.activation_field

    def strain_energy(self, F):
        r"""
        Strain-energy density function.

        .. math::

           \mathcal{W} = \mathcal{W}_1 + \mathcal{W}_{4f}
           + \mathcal{W}_{\mathrm{active}}

        where

        .. math::

           \mathcal{W}_{\mathrm{active}} =
           \begin{cases}
             0 & \text{if acitve strain} \\
             \gamma I_{4f} & \text{if active stress}
           \end{cases}


        :param F: Deformation gradient
        :type F: :py:class:`dolfin.Function`

        """

        # Invariants
        I1 = self.active.I1(F)
        I4f = self.active.I4(F)

        # Active stress
        Wactive = self.active.Wactive(F, diff=0)

        dim = get_dimesion(F)
        W1 = self.W_1(I1, diff=0, dim=dim)
        W4f = self.W_4(I4f, diff=0)

        W = W1 + W4f + Wactive

        return W

    def W_1(self, *args, **kwargs):
        return 0

    def W_4(self, *args, **kwargs):
        return 0

    def CauchyStress(self, F, p=None, deviatoric=False):

        I = kinematics.SecondOrderIdentity(F)

        F = dolfin.variable(F)
        # J = dolfin.variable(det(F))

        # First Piola Kirchoff
        if deviatoric:
            P = self.FirstPiolaStress(F, None)
        else:
            P = self.FirstPiolaStress(F, p)

        # Cauchy stress
        T = kinematics.InversePiolaTransform(P, F)

        return T

    def SecondPiolaStress(self, F, p=None, deviatoric=False, *args, **kwargs):

        I = kinematics.SecondOrderIdentity(F)

        f0 = self.f0
        f0f0 = dolfin.outer(f0, f0)

        I1 = dolfin.variable(self.active.I1(F))
        I4f = dolfin.variable(self.active.I4(F))

        Fe = self.active.Fe(F)
        Fa = self.active.Fa
        Ce = Fe.T * Fe

        # fe = Fe*f0
        # fefe = dolfin.outer(fe, fe)

        # Elastic volume ratio
        J = dolfin.variable(dolfin.det(Fe))
        # Active volume ration
        Ja = dolfin.det(Fa)

        dim = get_dimesion(F)
        Ce_bar = pow(J, -2.0 / float(dim)) * Ce

        w1 = self.W_1(I1, diff=1, dim=dim)
        w4f = self.W_4(I4f, diff=1)

        # Total Stress
        S_bar = Ja * (2 * w1 * I + 2 * w4f * f0f0) * dolfin.inv(Fa).T

        if self.is_isochoric:

            # Deviatoric
            Dev_S_bar = S_bar - (1.0 / 3.0) * dolfin.inner(S_bar, Ce_bar) * dolfin.inv(
                Ce_bar
            )

            S_mat = J ** (-2.0 / 3.0) * Dev_S_bar
        else:
            S_mat = S_bar

        # Volumetric
        if p is None or deviatoric:
            S_vol = dolfin.zero((dim, dim))
        else:
            psi_vol = self.compressibility(p, J)
            S_vol = J * dolfin.diff(psi_vol, J) * dolfin.inv(Ce)

        # Active stress
        wactive = self.active.Wactive(F, diff=1)
        eta = self.active.eta

        S_active = wactive * (f0f0 + eta * (I - f0f0))

        S = S_mat + S_vol + S_active

        return S

    def FirstPiolaStress(self, F, p=None, *args, **kwargs):

        I = kinematics.SecondOrderIdentity(F)
        F = dolfin.variable(F)

        # First Piola Kirchoff
        psi_iso = self.strain_energy(F)
        P_iso = dolfin.diff(psi_iso, F)

        if p is None:
            return P_iso
        else:
            J = dolfin.variable(dolfin.det(F))
            psi_vol = self.compressibility(p, J)
            P_vol = J * dolfin.diff(psi_vol, J) * dolfin.inv(F).T

            P = P_iso + P_vol

            return P



#!/usr/bin/env python
from dolfin import det, grad as Grad, inv, Identity, tr, inner

from .dolfin_utils import get_dimesion


# Strain tensors #####

# Second order identity tensor
def SecondOrderIdentity(F):
    """Return identity with same dimension as input
    """
    dim = get_dimesion(F)
    return Identity(dim)


def DeformationGradient(u, isochoric=False):
    """Return deformation gradient from displacement.
    """
    I = SecondOrderIdentity(u)
    F = I + Grad(u)
    if isochoric:
        return IsochoricDeformationGradient(F)
    else:
        return F


def IsochoricDeformationGradient(F):
    """Return the isochoric deformation gradient
    """
    J = Jacobian(F)
    dim = get_dimesion(F)
    return pow(J, -1.0 / float(dim)) * F


def Jacobian(F):
    """Determinant of the deformation gradient
    """
    return det(F)


def EngineeringStrain(F, isochoric=False):
    """Equivalent of engineering strain
    """
    I = SecondOrderIdentity(F)
    if isochoric:
        return IsochoricDeformationGradient(F) - I
    else:
        return F - I


def GreenLagrangeStrain(F, isochoric=False):
    """Green-Lagrange strain tensor
    """
    I = SecondOrderIdentity(F)
    C = RightCauchyGreen(F, isochoric)
    return 0.5 * (C - I)


def LeftCauchyGreen(F, isochoric=False):
    """Left Cauchy-Green tensor
    """
    if isochoric:
        F_ = IsochoricDeformationGradient(F)
    else:
        F_ = F

    return F_ * F_.T


def RightCauchyGreen(F, isochoric=False):
    """Left Cauchy-Green tensor
    """
    if isochoric:
        F_ = IsochoricDeformationGradient(F)
    else:
        F_ = F

    return F_.T * F_


def EulerAlmansiStrain(F, isochoric=False):
    """Euler-Almansi strain tensor
    """
    I = SecondOrderIdentity(F)
    b = LeftCauchyGreen(F, isochoric)
    return 0.5 * (I - inv(b))


# Invariants #####
class Invariants(object):
    def __init__(self, isochoric=True, *args):
        self._isochoric = isochoric

    @property
    def is_isochoric(self):
        return self._isochoric

    def _I1(self, F):

        C = RightCauchyGreen(F, self._isochoric)
        I1 = tr(C)
        return I1

    def _I2(self, F):
        C = RightCauchyGreen(F, self._isochoric)
        return 0.5 * (self._I1(F) * self._I1(F) - tr(C * C))

    def _I3(self, F):
        C = RightCauchyGreen(F, self._isochoric)
        return det(C)

    def _I4(self, F, a0):

        C = RightCauchyGreen(F, self._isochoric)
        I4 = inner(C * a0, a0)
        return I4

    def _I5(self, F, a0):
        C = RightCauchyGreen(F, self._isochoric)
        I5 = inner(C * a0, C * a0)
        return I5

    def _I6(self, F, a0):
        return self._I4(F, a0)

    def _I7(self, F, a0):
        return self._I5(F, a0)

    def _I8(self, u, a0, b0):
        C = RightCauchyGreen(F, self._isochoric)
        I8 = inner(C * a0, C * b0)
        return I8


# Transforms #####
# Pull-back of a two-tensor from the current to the reference
# configuration
def PiolaTransform(A, F):
    J = Jacobian(F)
    B = J * A * inv(F).T
    return B


# Push-forward of a two-tensor from the reference to the current
# configuration
def InversePiolaTransform(A, F):
    J = Jacobian(F)
    B = (1 / J) * A * F.T
    return B



#!/usr/bin/env python
import os
import logging
import numpy as np
import dolfin
from . import numpy_mpi
from .utils import make_logger, mpi_comm_world
from .io_utils import check_and_delete
from .iterate import iterate, logger as logger_it
from .dolfin_utils import get_pressure

from . import parameters

logger = make_logger(__name__, parameters["log_level"])


class ResidualCalculator(object):
    def __init__(self, mesh):

        try:
            from scipy import spatial

        except ImportError:
            logger.warning(
                (
                    "Scipy is not install. Residual in the unloading "
                    "algorithm cannot be computed. Please install "
                    'scipy "pip install scipy" if you want to compute '
                    "the residual"
                )
            )
            self.bbtree = None

        else:

            self.mesh = mesh
            d = self.mesh.topology().dim()
            local_points = [v.point() for v in dolfin.vertices(self.mesh)]
            coords = [(p.x(), p.y(), p.z()) for p in local_points]

            # FIXME
            coords = numpy_mpi.gather_broadcast(np.array(coords).flatten())
            coords.resize(int(len(coords) / d), d)

            self.bbtree = spatial.KDTree(coords)

    def calculate_residual(self, mesh2):
        boundmesh = dolfin.BoundaryMesh(mesh2, "exterior")
        d = max(
            [
                self.bbtree.query(dolfin.Vertex(boundmesh, v_idx).point().array())[0]
                for v_idx in range(boundmesh.num_vertices())
            ]
        )

        return dolfin.MPI.max(mpi_comm_world(), d)


def save(obj, h5name, name, h5group=""):
    """
    Save object to and HDF file.

    Parameters
    ----------

    obj : dolfin.Mesh or dolfin.Function
        The object you want to save
    name : str
        Name of the object
    h5group : str
        The folder you want to save the object
        withing the HDF file. Default: ''

    """
    group = "/".join([h5group, name])
    file_mode = "a" if os.path.isfile(h5name) else "w"

    if os.path.isfile(h5name):
        check_and_delete(h5name, group)
        file_mode = "a"
    else:
        file_mode = "w"

    logger.debug("Save {0} to {1}:{2}/{0}".format(name, h5name, h5group))

    if isinstance(obj, dolfin.Function) and obj.ufl_element().family() == "Quadrature":

        quad_to_xdmf(obj, h5name, group, file_mode)

    else:
        with dolfin.HDF5File(mpi_comm_world(), h5name, file_mode) as h5file:
            h5file.write(obj, group)


def quad_to_xdmf(obj, h5name, h5group="", file_mode="w"):

    V = obj.function_space()
    gx, gy, gz = obj.split(deepcopy=True)

    W = V.sub(0).collapse()
    # FIXME
    coords_tmp = numpy_mpi.gather_broadcast(W.tabulate_dof_coordinates())
    coords = coords_tmp.reshape((-1, 3))
    u = numpy_mpi.gather_vector(gx.vector())
    v = numpy_mpi.gather_vector(gy.vector())
    w = numpy_mpi.gather_vector(gz.vector())
    vecs = np.array([u, v, w]).T
    from ..io_utils import open_h5py, parallel_h5py

    with open_h5py(h5name) as h5file:

        if not parallel_h5py:
            if mpi_comm_world().rank == 0:

                h5file.create_dataset("/".join([h5group, "coordinates"]), data=coords)
                h5file.create_dataset("/".join([h5group, "vector"]), data=vecs)
        else:
            h5file.create_dataset("/".join([h5group, "coordinates"]), data=coords)
            h5file.create_dataset("/".join([h5group, "vector"]), data=vecs)


def inflate_to_pressure(target_pressure, problem, ntries=5, n=2, annotate=False):

    logger.debug("\nInflate geometry to p = {} kPa".format(target_pressure))
    pressure = get_pressure(problem)
    solve(target_pressure, problem, pressure, ntries, n, annotate)

    return problem.get_displacement(annotate=annotate)


def print_volumes(geometry, logger=logger, txt="original", u=None):

    logger.info(
        ("\nLV Volume of {} geometry = {:.3f} ml" "").format(
            txt, geometry.cavity_volume(chamber="lv")
        )
    )
    if geometry.is_biv:
        logger.info(
            ("RV Volume of {} geometry = {:.3f} ml" "").format(
                txt, geometry.cavity_volume(chamber="rv")
            )
        )


def solve(target_pressure, problem, pressure, ntries=5, n=2, annotate=False):

    if "adjoint" in dolfin.parameters:
        dolfin.parameters["adjoint"]["stop_annotating"] = True

    level = logger_it.level
    logger_it.setLevel(logging.WARNING)

    try:
        iterate(problem, pressure, target_pressure)

    except:
        from IPython import embed

        embed()
        exit()

    if annotate and "adjoint" in dolfin.parameters:
        # Only record the last solve, otherwise it becomes too
        # expensive memorywise.
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        problem.solve()

    logger_it.setLevel(level)
    w = problem.state.copy(deepcopy=True)
    return w


def update_material_parameters(material_parameters, mesh, merge_control_str=""):

    from ..setup_optimization import RegionalParameter, merge_control

    new_matparams = {}
    for k, v in material_parameters.iteritems():
        if isinstance(v, RegionalParameter):
            geo = lambda: None
            geo.mesh = mesh
            geo.sfun = dolfin.MeshFunction("size_t", mesh, 3, mesh.domains())
            sfun = merge_control(geo, merge_control_str)

            v_new = RegionalParameter(sfun)
            v_arr = numpy_mpi.gather_vector(v.vector())

            numpy_mpi.assign_to_vector(v_new.vector(), v_arr)
            new_matparams[k] = v_new

        elif isinstance(v, dolfin.Function):
            v_new = dolfin.Function(
                dolfin.FunctionSpace(mesh, v.function_space().ufl_element())
            )
            v_arr = numpy_mpi.gather_vector(v.vector())
            numpy_mpi.assign_to_vector(v_new.vector(), v_arr)
            new_matparams[k] = v_new

        else:
            new_matparams[k] = v

    return new_matparams


def load_opt_target(h5name, h5group, key="volume", data="simulated"):

    with h5py.File(h5name) as f:
        vols = [a[:][0] for a in f[h5group]["passive_inflation"][key][data].values()]

    return vols


def save_unloaded_geometry(new_geometry, h5name, h5group, backward_displacement=None):

    fields = ["fiber", "sheet", "sheet_normal"]
    local_basis = ["circumferential", "radial", "longitudinal"]

    new_fields = []
    for fattr in fields:
        if hasattr(new_geometry, fattr) and getattr(new_geometry, fattr) is not None:
            f = getattr(new_geometry, fattr).copy()
            f.rename(fattr, "microstructure")
            new_fields.append(f)

            new_local_basis = []
    for fattr in local_basis:
        if hasattr(new_geometry, fattr) and getattr(new_geometry, fattr) is not None:
            f = getattr(new_geometry, fattr).copy()
            f.rename(fattr, "local_basis_function")
            new_local_basis.append(f)

    logger.debug("Save geometry to {}:{}".format(h5name, h5group))

    if backward_displacement:
        other_functions = {"backward_displacement": backward_displacement}
    else:
        other_functions = {}

    from mesh_generation.mesh_utils import save_geometry_to_h5

    save_geometry_to_h5(
        new_geometry.mesh,
        h5name,
        h5group,
        new_geometry.markers,
        new_fields,
        new_local_basis,
        other_functions=other_functions,
    )


def continuation_step(params, it_, paramvec):

    # Use data from the two prevoious steps and continuation
    # to get a good next gues
    values = []
    vols = []

    v_target = load_opt_target(params["sim_file"], "0", "volume", "target")
    for it in range(it_):
        p_tmp = dolfin.Function(paramvec.function_space())
        load_material_parameter(params["sim_file"], str(it), p_tmp)

        values.append(numpy_mpi.gather_vector(p_tmp.vector()))

        v = load_opt_target(params["sim_file"], str(it), "volume", "simulated")
        vols.append(v)

    ed_vols = np.array(vols).T[-1]
    # Make continuation approximation
    delta = (v_target[-1] - ed_vols[-2]) / (ed_vols[-1] - ed_vols[-2])
    a_cont = (1 - delta) * values[-2] + delta * values[-1]
    a_prev = values[-1]

    # Make sure next step is not to far away
    if hasattr(a_cont, "__len__"):

        a_next = np.array(
            [
                min(max(a_cont[i], a_prev[i] / 2), a_prev[i] * 2)
                for i in range(len(a_cont))
            ]
        )

        # Just make sure that we are within the given bounds
        a = np.array(
            [
                min(
                    max(a_next[i], params["Optimization_parameters"]["matparams_min"]),
                    params["Optimization_parameters"]["matparams_max"],
                )
                for i in range(len(a_cont))
            ]
        )

    else:

        a_next = min(max(a_cont, a_prev / 2), a_prev * 2)

        # Just make sure that we are within the given bounds
        a = min(
            max(a_next, params["Optimization_parameters"]["matparams_min"]),
            params["Optimization_parameters"]["matparams_max"],
        )

    numpy_mpi.assign_to_vector(paramvec.vector(), a)


def load_material_parameter(h5name, h5group, paramvec):
    logger.info("Load {}:{}".format(h5name, h5group))
    group = "/".join([h5group, "passive_inflation", "optimal_control"])
    with dolfin.HDF5File(mpi_comm_world(), h5name, "r") as h5file:
        h5file.read(paramvec, group)



import os

import dolfin
import h5py
import numpy as np

try:
    from dolfin_adjoint import Function
except ImportError:
    from dolfin import Function

# For newer versions of dolfin_adjoint
try:
    from dolfin_adjoint import Mesh
except ImportError:
    from dolfin import Mesh

from . import io_utils, parameters
from .utils import DOLFIN_VERSION_MAJOR, make_logger, mpi_comm_world

logger = make_logger(__name__, parameters["log_level"])


def move(mesh, u, factor=1.0):
    """
    Move mesh according to some displacement times some factor
    """
    W = dolfin.VectorFunctionSpace(u.function_space().mesh(), "CG", 1)

    # Use interpolation for now. It is the only thing that makes sense
    u_int = dolfin.interpolate(u, W)

    u0 = dolfin.Function(W)
    # arr = factor * numpy_mpi.gather_vector(u_int.vector())
    # numpy_mpi.assign_to_vector(u0.vector(), arr)
    u0.vector()[:] = factor * u_int.vector()
    V = dolfin.VectorFunctionSpace(mesh, "CG", 1)
    U = dolfin.Function(V)
    U.vector()[:] = u0.vector()
    # numpy_mpi.assign_to_vector(U.vector(), arr)

    dolfin.ALE.move(mesh, U)

    return u0


def get_fiber_markers(mesh_type="lv"):
    """
    Get the markers for the mesh.
    This is the default markers for fiberrules.

    :param str mesh_type: type of mesh, 'lv' or 'biv'
    :returns: The markers
    :rtype: dict

    """
    if mesh_type == "lv":

        return {
            "BASE": 10,
            "ENDO": 30,
            "EPI": 40,
            "WALL": 50,
            "ENDORING": 300,
            "EPIRING": 400,
            "WALL": 50,
        }

    elif mesh_type == "biv":

        return {
            "BASE": 10,
            "ENDO_RV": 20,
            "ENDO_LV": 30,
            "EPI": 40,
            "ENDORING_RV": 200,
            "ENDORING_LV": 300,
            "EPIRING": 400,
            "WALL": 50,
        }


def get_markers(mesh_type="lv"):

    assert mesh_type in ["lv", "biv"]

    fiber_markers = get_fiber_markers(mesh_type)

    markers = {}
    markers["NONE"] = (0, 3)

    markers["BASE"] = (fiber_markers["BASE"], 2)
    markers["EPI"] = (fiber_markers["EPI"], 2)
    markers["EPIRING"] = (fiber_markers["EPIRING"], 1)

    if mesh_type == "lv":

        markers["ENDO"] = (fiber_markers["ENDO"], 2)

        markers["ENDORING"] = (fiber_markers["ENDORING"], 1)

    else:

        markers["ENDO_RV"] = (fiber_markers["ENDO_RV"], 2)
        markers["ENDO_LV"] = (fiber_markers["ENDO_LV"], 2)

        markers["ENDORING_RV"] = (fiber_markers["ENDORING_RV"], 1)
        markers["ENDORING_LV"] = (fiber_markers["ENDORING_LV"], 1)

    return markers


def get_geometric_dimension(mesh):
    try:
        return mesh.geometric_dimension()
    except AttributeError:
        return mesh.geometry().dim()


def load_geometry_from_h5(
    h5name,
    h5group="",
    fendo=None,
    fepi=None,
    include_sheets=True,
    comm=mpi_comm_world(),
):
    """Load geometry and other mesh data from
    a h5file to an object.
    If the file contains muliple fiber fields
    you can spefify the angles, and if the file
    contais sheets and cross-sheets this can also
    be included

    :param str h5name: Name of the h5file
    :param str h5group: The group within the file
    :param int fendo: Helix fiber angle (endocardium) (if available)
    :param int fepi: Helix fiber angle (epicardium) (if available)
    :param bool include_sheets: Include sheets and cross-sheets
    :returns: An object with geometry data
    :rtype: object

    """

    logger.info("\nLoad mesh from h5")
    # Set default groups
    ggroup = "{}/geometry".format(h5group)
    mgroup = "{}/mesh".format(ggroup)
    lgroup = "{}/local basis functions".format(h5group)
    fgroup = "{}/microstructure/".format(h5group)

    if not os.path.isfile(h5name):
        raise IOError("File {} does not exist".format(h5name))

    # Check that the given file contains
    # the geometry in the given h5group
    if not io_utils.check_h5group(h5name, mgroup, delete=False, comm=comm):
        msg = ("Warning!\nGroup: '{}' does not exist in file:" "\n{}").format(
            mgroup, h5name
        )

        with h5py.File(h5name) as h:
            keys = h.keys()
        msg += "\nPossible values for the h5group are {}".format(keys)
        raise IOError(msg)

    # Create a dummy object for easy parsing
    class Geometry(object):
        pass

    geo = Geometry()

    with dolfin.HDF5File(comm, h5name, "r") as h5file:

        # Load mesh
        mesh = Mesh(comm)
        io_utils.read_h5file(h5file, mesh, mgroup, True)
        geo.mesh = mesh

        # Get mesh functions
        for dim, attr in enumerate(["vfun", "efun", "ffun", "cfun"]):

            if dim > get_geometric_dimension(mesh):
                setattr(geo, attr, None)
                continue

            dgroup = "{}/mesh/meshfunction_{}".format(ggroup, dim)
            mf = dolfin.MeshFunction("size_t", mesh, dim, mesh.domains())

            if h5file.has_dataset(dgroup):
                io_utils.read_h5file(h5file, mf, dgroup)
            setattr(geo, attr, mf)

        load_local_basis(h5file, lgroup, mesh, geo)
        load_microstructure(h5file, fgroup, mesh, geo, include_sheets)

        # Load the boundary markers
        markers = load_markers(h5file, mesh, ggroup, dgroup)
        geo.markers = markers

        origmeshgroup = "{}/original_geometry".format(h5group)
        if h5file.has_dataset(origmeshgroup):
            original_mesh = Mesh(comm)
            io_utils.read_h5file(h5file, original_mesh, origmeshgroup, True)
            setattr(geo, "original_geometry", original_mesh)

    for attr in ["f0", "s0", "n0", "r0", "c0", "l0", "cfun", "vfun", "efun", "ffun"]:
        if not hasattr(geo, attr):
            setattr(geo, attr, None)

    return geo


def load_markers(h5file, mesh, ggroup, dgroup):
    try:
        markers = {}
        for dim in range(mesh.ufl_domain().topological_dimension() + 1):
            for key_str in ["domain", "meshfunction"]:
                dgroup = "{}/mesh/{}_{}".format(ggroup, key_str, dim)

                # If dataset is not present
                if not h5file.has_dataset(dgroup):
                    continue

                def get_attributes():
                    if DOLFIN_VERSION_MAJOR >= 2018:
                        return h5file.attributes(dgroup).list_attributes()
                    else:
                        return h5file.attributes(dgroup).str().strip().split(" ")

                for aname in get_attributes():
                    if aname.startswith("marker_name"):

                        name = aname.rsplit("marker_name_")[-1]
                        marker = h5file.attributes(dgroup)[
                            "marker_name_{}".format(name)
                        ]
                        markers[name] = (int(marker), dim)

    except Exception as ex:
        logger.info("Unable to load makers")
        logger.info(ex)
        markers = get_markers()

    return markers


def setup_fiber_parameters():
    fiber_params = dolfin.Parameters("Fibers")
    fiber_params.add("fiber_space", "Quadrature_4")
    fiber_params.add("include_sheets", False)

    fiber_params.add("fiber_angle_epi", -60)
    fiber_params.add("fiber_angle_endo", 60)
    fiber_params.add("sheet_angle_epi", 0)
    fiber_params.add("sheet_angle_endo", 0)
    return fiber_params


def full_arctangent(x, y):
    t = np.arctan2(x, y)
    if t < 0:
        return t + 2 * np.pi
    else:
        return t


def cartesian_to_prolate_ellipsoidal(x, y, z, a):

    b1 = np.sqrt((x + a) ** 2 + y ** 2 + z ** 2)
    b2 = np.sqrt((x - a) ** 2 + y ** 2 + z ** 2)

    sigma = 1 / (2.0 * a) * (b1 + b2)
    tau = 1 / (2.0 * a) * (b1 - b2)
    phi = full_arctangent(z, y)
    mu = np.arccosh(sigma)
    nu = np.arccos(tau)
    return mu, nu, phi


def prolate_ellipsoidal_to_cartesian(mu, nu, phi, a):
    x = a * np.cosh(mu) * np.cos(nu)
    y = a * np.sinh(mu) * np.sin(nu) * np.cos(phi)
    z = a * np.sinh(mu) * np.sin(nu) * np.sin(phi)
    return x, y, z


def fill_coordinates_ec(i, e_c_x, e_c_y, e_c_z, coord, foci):
    norm = dolfin.sqrt(coord[1] ** 2 + coord[2] ** 2)
    if not dolfin.near(norm, 0):
        e_c_y.vector()[i] = -coord[2] / norm
        e_c_z.vector()[i] = coord[1] / norm
    else:
        # We are at the apex where crl system doesn't make sense
        # So just pick something.
        e_c_y.vector()[i] = 1
        e_c_z.vector()[i] = 0


def fill_coordinates_el(i, e_c_x, e_c_y, e_c_z, coord, foci):

    norm = dolfin.sqrt(coord[1] ** 2 + coord[2] ** 2)
    if not dolfin.near(norm, 0):
        mu, nu, phi = cartesian_to_prolate_ellipsoidal(*(coord.tolist() + [foci]))
        x, y, z = prolate_ellipsoidal_to_cartesian(mu, nu + 0.01, phi, foci)
        r = np.array([coord[0] - x, coord[1] - y, coord[2] - z])
        e_r = r / np.linalg.norm(r)
        e_c_x.vector()[i] = e_r[0]
        e_c_y.vector()[i] = e_r[1]
        e_c_z.vector()[i] = e_r[2]
    else:
        e_c_y.vector()[i] = 0
        e_c_z.vector()[i] = 1


def calc_cross_products(e1, e2, VV):
    e_crossed = dolfin.Function(VV)

    e1_arr = e1.vector().get_local().reshape((-1, 3))
    e2_arr = e2.vector().get_local().reshape((-1, 3))

    crosses = []
    for c1, c2 in zip(e1_arr, e2_arr):
        crosses.extend(np.cross(c1, c2.tolist()))

    e_crossed.vector()[:] = np.array(crosses)[:]
    return e_crossed


def make_unit_vector(V, VV, dofs_x, fill_coordinates, foc=None):
    e_c_x = dolfin.Function(V)
    e_c_y = dolfin.Function(V)
    e_c_z = dolfin.Function(V)

    for i, coord in enumerate(dofs_x):
        fill_coordinates(i, e_c_x, e_c_y, e_c_z, coord, foc)

    e = dolfin.Function(VV)

    fa = [dolfin.FunctionAssigner(VV.sub(i), V) for i in range(3)]
    for i, e_c_comp in enumerate([e_c_x, e_c_y, e_c_z]):
        fa[i].assign(e.split()[i], e_c_comp)
    return e


def make_crl_basis(mesh, foc):
    """
    Makes the crl  basis for the LV mesh (prolate ellipsoidal)
    with prespecified focal length.
    """

    VV = dolfin.VectorFunctionSpace(mesh, "CG", 1)
    V = dolfin.FunctionSpace(mesh, "CG", 1)

    if DOLFIN_VERSION_MAJOR > 1.6:
        dofs_x = V.tabulate_dof_coordinates().reshape((-1, mesh.geometry().dim()))
    else:
        dm = V.dofmap()
        dofs_x = dm.tabulate_all_coordinates(mesh).reshape((-1, mesh.geometry().dim()))

    e_c = make_unit_vector(V, VV, dofs_x, fill_coordinates_ec)
    e_l = make_unit_vector(V, VV, dofs_x, fill_coordinates_el, foc)
    e_r = calc_cross_products(e_c, e_l, VV)

    e_c.rename("c0", "local_basis_function")
    e_r.rename("r0", "local_basis_function")
    e_l.rename("l0", "local_basis_function")

    return e_c, e_r, e_l


def make_BiV_crl_basis(mesh):

    c0 = get_circ_field(mesh)
    l0 = get_long_field(mesh, "biv")
    r0 = calc_cross_products(c0, l0, c0.function_space())

    return c0, r0, l0


def generate_local_basis_functions(mesh, focal_point, mesh_type="lv"):

    assert mesh_type in ["lv", "biv"], "Excepted mesh_type to be 'lv' or 'biv'"

    if mesh_type == "biv":

        try:
            c0, r0, l0 = make_BiV_crl_basis(mesh)

        except ImportError:
            pass
        else:
            return c0, r0, l0

    # Make basis functions
    c0, r0, l0 = make_crl_basis(mesh, focal_point)

    return c0, r0, l0


def fibers(mesh, fiber_endo=60, fiber_epi=-60):

    fiber_params = setup_fiber_parameters()
    fiber_params["fiber_angle_endo"] = fiber_endo
    fiber_params["fiber_angle_epi"] = fiber_epi

    f0 = generate_fibers(mesh, fiber_params)[0]
    return f0


def get_circ_field(mesh):

    fiber_params = dolfin.Parameters("Fibers")
    fiber_params.add("fiber_space", "Quadrature_4")
    # fiber_params.add("fiber_space", "Lagrange_1")
    fiber_params.add("include_sheets", False)

    # Parameter set from Bayer et al.
    fiber_params.add("fiber_angle_epi", 0)
    fiber_params.add("fiber_angle_endo", 0)
    fiber_params.add("sheet_angle_epi", 0)
    fiber_params.add("sheet_angle_endo", 0)

    f0 = generate_fibers(mesh, fiber_params)[0]
    f0.rename("circumferential", "local_basis_function")

    return f0


def get_long_field(mesh, mesh_type="biv"):

    fiber_params = setup_fiber_parameters()
    fiber_params["fiber_angle_endo"] = -90
    fiber_params["fiber_angle_epi"] = -90

    if mesh_type == "biv":
        # We need to set the markers for then LV and RV facets
        ffun = dolfin.MeshFunction("size_t", mesh, 2, mesh.domains())

        markers = get_fiber_markers("biv")
        # Mark the mesh with same markers on the LV and RV before
        # running the LDRB algorithm
        markers["ENDO_RV"] = markers["ENDO_LV"]
        for facet in dolfin.facets(mesh):
            if ffun[facet] != 0:
                mesh.domains().set_marker((facet.index(), markers[ffun[facet]]), 2)

    f0 = generate_fibers(mesh, fiber_params)[0]
    f0.rename("longitudinal", "local_basis_function")

    if mesh_type == "biv":
        # Put the correct markers
        markers = get_fiber_markers("biv")
        for facet in dolfin.facets(mesh):
            if ffun[facet] != 0:
                mesh.domains().set_marker((facet.index(), markers[ffun[facet]]), 2)
    return f0


def generate_fibers(mesh, fiber_params, ffun=None):
    """
    Generate fibers on mesh based on provided parameters.
    It is not recomemmended to use this function.
    Use the `ldrb` package directly instead. This function is
    mainly used to ensure version compatability.
    """

    try:
        from ldrb import dolfin_ldrb
    except ImportError:
        msg = (
            '"ldrb" package not found. Please go to '
            "https://github.com/finsberg/ldrb to see how you can get it!"
        )
        print(msg)
        raise ImportError(msg)

    if isinstance(fiber_params, dolfin.Parameters):
        p = fiber_params.to_dict()
    else:
        p = fiber_params

    angles = dict(
        alpha_endo_lv=p.get("fiber_angle_endo"), alpha_epi_lv=p.get("fiber_angle_epi")
    )
    for name, a in angles.items():
        if isinstance(a, dolfin.cpp.parameter.Parameter):
            angles[name] = a.value()

    return dolfin_ldrb(mesh, ffun=ffun, **angles)


def generate_fibers_old(mesh, fiber_params):

    try:
        from fiberrules import dolfin_fiberrules
    except ImportError as ex:
        logger.error("Unable to generate fibers without fiberrules package")
        logger.error("This package is protected by copyright")
        raise ex

    logger.info("\nGENERATING FIBERS ...")

    fiber_space_name = fiber_params["fiber_space"]

    assert (
        len(fiber_space_name.split("_")) == 2
    ), "expected fiber_space_name in 'FamilyName_Degree' format"

    family, degree = fiber_space_name.split("_")

    if DOLFIN_VERSION_MAJOR > 1.6:
        el = dolfin.FiniteElement(
            family=family,
            cell=mesh.ufl_cell(),
            degree=int(degree),
            quad_scheme="default",
        )
        fiber_space = dolfin.FunctionSpace(mesh, el)
    else:
        fiber_space = dolfin.FunctionSpace(mesh, family, int(degree))

    if family == "Quadrature":
        dolfin.parameters["form_compiler"]["quadrature_degree"] = int(degree)
        if DOLFIN_VERSION_MAJOR > 2016:
            dolfin.parameters["form_compiler"]["representation"] = "quadrature"

    # There are some strange shifting in the fiberrults angles
    fiber_angle_epi = 90 - (-fiber_params["fiber_angle_epi"])
    fiber_angle_endo = 90 - (fiber_params["fiber_angle_endo"])
    sheet_angle_endo = fiber_params.to_dict().get("sheet_angle_endo", 0)
    sheet_angle_epi = fiber_params.to_dict().get("sheet_angle_epi", 0)

    microstructures = dolfin_fiberrules(
        mesh,
        fiber_space,
        fiber_angle_epi,
        fiber_angle_endo,
        sheet_angle_epi,
        sheet_angle_endo,
    )

    if DOLFIN_VERSION_MAJOR > 2016:
        dolfin.parameters["form_compiler"]["representation"] = "uflacs"

    microstructures[0].rename(
        "fiber",
        "epi{}_endo{}".format(
            fiber_params["fiber_angle_epi"], fiber_params["fiber_angle_endo"]
        ),
    )
    fields = [microstructures[0]]

    if fiber_params["include_sheets"]:
        microstructures[1].rename(
            "sheet", "epi{}_endo{}".format(sheet_angle_epi, sheet_angle_endo)
        )
        microstructures[2].rename(
            "cross_sheet",
            "fepi{}_fendo{}_sepi{}_sendo{}".format(
                fiber_params["fiber_angle_epi"],
                fiber_params["fiber_angle_endo"],
                sheet_angle_epi,
                sheet_angle_endo,
            ),
        )
        fields.append(microstructures[1])
        fields.append(microstructures[2])

    return fields


def load_local_basis(h5file, lgroup, mesh, geo):

    if h5file.has_dataset(lgroup):
        # Get local bais functions
        local_basis_attrs = h5file.attributes(lgroup)
        lspace = local_basis_attrs["space"]
        family, order = lspace.split("_")

        namesstr = local_basis_attrs["names"]
        names = namesstr.split(":")

        if DOLFIN_VERSION_MAJOR > 1.6:
            elm = dolfin.VectorElement(
                family=family,
                cell=mesh.ufl_cell(),
                degree=int(order),
                quad_scheme="default",
            )
            V = dolfin.FunctionSpace(mesh, elm)
        else:
            V = dolfin.VectorFunctionSpace(mesh, family, int(order))

        for name in names:
            lb = Function(V, name=name)

            io_utils.read_h5file(h5file, lb, lgroup + "/{}".format(name))
            setattr(geo, name, lb)
    else:
        setattr(geo, "circumferential", None)
        setattr(geo, "radial", None)
        setattr(geo, "longitudinal", None)


def load_microstructure(h5file, fgroup, mesh, geo, include_sheets=True):

    if h5file.has_dataset(fgroup):
        # Get fibers
        fiber_attrs = h5file.attributes(fgroup)
        fspace = fiber_attrs["space"]
        if fspace is None:
            # Assume quadrature 4
            family = "Quadrature"
            order = 4
        else:
            family, order = fspace.split("_")

        namesstr = fiber_attrs["names"]
        if namesstr is None:
            names = ["fiber"]
        else:
            names = namesstr.split(":")

        # Check that these fibers exists
        for name in names:
            fsubgroup = fgroup + "/{}".format(name)
            if not h5file.has_dataset(fsubgroup):
                msg = ("H5File does not have dataset {}").format(fsubgroup)
                logger.warning(msg)

        if DOLFIN_VERSION_MAJOR > 1.6:
            elm = dolfin.VectorElement(
                family=family,
                cell=mesh.ufl_cell(),
                degree=int(order),
                quad_scheme="default",
            )
            V = dolfin.FunctionSpace(mesh, elm)
        else:
            V = dolfin.VectorFunctionSpace(mesh, family, int(order))

        attrs = ["f0", "s0", "n0"]
        for i, name in enumerate(names):
            func = Function(V, name=name)
            fsubgroup = fgroup + "/{}".format(name)

            io_utils.read_h5file(h5file, func, fsubgroup)

            setattr(geo, attrs[i], func)


def save_geometry_to_h5(
    mesh,
    h5name,
    h5group="",
    markers=None,
    fields=None,
    local_basis=None,
    meshfunctions=None,
    comm=None,
    other_functions=None,
    other_attributes=None,
    overwrite_file=False,
    overwrite_group=True,
):
    """
    Save geometry and other geometrical functions to a HDF file.

    Parameters
    ----------
    mesh : :class:`dolfin.mesh`
        The mesh
    h5name : str
        Path to the file
    h5group : str
        Folder within the file. Default is "" which means in
        the top folder.
    markers : dict
        A dictionary with markers. See `get_markers`.
    fields : list
        A list of functions for the microstructure
    local_basis : list
        A list of functions for the crl basis
    meshfunctions : dict
        A dictionary with keys being the dimensions the the values
        beeing the meshfunctions.
    comm : :class:`dolfin.MPI`
        MPI communicator
    other_functions : dict
        Dictionary with other functions you want to save
    other_attributes: dict
        Dictionary with other attributes you want to save
    overwrite_file : bool
        If true, and the file exists, the file will be overwritten (default: False)
    overwrite_group : bool
        If true and h5group exist, the group will be overwritten.

    """

    logger.info("\nSave mesh to h5")
    assert isinstance(mesh, dolfin.Mesh)
    if comm is None:
        comm = mesh.mpi_comm()
    file_mode = "a" if os.path.isfile(h5name) and not overwrite_file else "w"

    # IF we should append the file but overwrite the group we need to
    # check that the group does not exist. If so we need to open it in
    # h5py and delete it.
    if file_mode == "a" and overwrite_group and h5group != "":
        io_utils.check_h5group(h5name, h5group, delete=True, comm=comm)

    with dolfin.HDF5File(comm, h5name, file_mode) as h5file:

        # Save mesh
        ggroup = "{}/geometry".format(h5group)
        mgroup = "{}/mesh".format(ggroup)
        h5file.write(mesh, mgroup)

        for dim in range(get_geometric_dimension(mesh) + 1):

            if meshfunctions is not None and dim in meshfunctions:
                mf = meshfunctions[dim]
            else:
                mf = dolfin.MeshFunction("size_t", mesh, dim, mesh.domains())

            save_mf = dolfin.MPI.max(comm, len(set(mf.array()))) > 1

            if save_mf:
                dgroup = "{}/mesh/meshfunction_{}".format(ggroup, dim)
                h5file.write(mf, dgroup)

        if markers is not None:
            # Save the boundary markers
            for name, (marker, dim) in markers.items():

                for key_str in ["domain", "meshfunction"]:
                    dgroup = "{}/mesh/{}_{}".format(ggroup, key_str, dim)

                    if h5file.has_dataset(dgroup):
                        aname = "marker_name_{}".format(name)
                        h5file.attributes(dgroup)[aname] = marker

        if local_basis is not None:
            # Save local basis functions
            lgroup = "{}/local basis functions".format(h5group)
            save_local_basis(h5file, lgroup, local_basis)

        if fields is not None:
            # Save fiber field
            fgroup = "{}/microstructure".format(h5group)
            save_fields(h5file, fgroup, fields)

        if other_functions is not None:
            for k, fun in other_functions.items():
                fungroup = "/".join([h5group, k])
                h5file.write(fun, fungroup)

            if isinstance(fun, dolfin.Function):
                elm = fun.function_space().ufl_element()
                family, degree, vsize = elm.family(), elm.degree(), elm.value_size()
                fspace = "{}_{}".format(family, degree)
                h5file.attributes(fungroup)["space"] = fspace
                h5file.attributes(fungroup)["value_size"] = vsize

        if other_attributes is not None:
            for k, v in other_attributes.iteritems():
                if isinstance(v, str) and isinstance(k, str):
                    h5file.attributes(h5group)[k] = v
                else:
                    logger.warning("Invalid attribute {} = {}".format(k, v))

    logger.info("Geometry saved to {}".format(h5name))


def save_fields(h5file, fgroup, fields):
    """Save mictrostructure to h5file

    Parameters
    ----------
    h5file : dolfin.HDF5File
        The file to write to
    fgroup : str
        The folder inside the file to write to
    fields : list
        List of dolfin functions to write
    """
    names = []
    for field in fields:
        try:
            label = field.label() if field.label().rfind("a Function") == -1 else ""
        except AttributeError:
            label = field.name()
        name = "_".join(filter(None, [str(field), label]))
        fsubgroup = "{}/{}".format(fgroup, name)
        h5file.write(field, fsubgroup)
        h5file.attributes(fsubgroup)["name"] = field.name()
        names.append(name)

    elm = field.function_space().ufl_element()
    family, degree = elm.family(), elm.degree()
    fspace = "{}_{}".format(family, degree)
    h5file.attributes(fgroup)["space"] = fspace
    h5file.attributes(fgroup)["names"] = ":".join(names)


def save_local_basis(h5file, lgroup, local_basis):
    """Save local basis functions

    Parameters
    ----------
    h5file : dolfin.HDF5File
        The file to write to
    lgroup : str
        Folder inside file to store the local basis
    local_basis : list
        List of dolfin functions with local basis functions
    """
    names = []
    for basis in local_basis:
        h5file.write(basis, lgroup + "/{}".format(basis.name()))
        names.append(basis.name())

    elm = basis.function_space().ufl_element()
    family, degree = elm.family(), elm.degree()
    lspace = "{}_{}".format(family, degree)
    h5file.attributes(lgroup)["space"] = lspace
    h5file.attributes(lgroup)["names"] = ":".join(names)


def mark_strain_regions(mesh, foc=None, nsectors=(6, 6, 4, 1), mark_mesh=True):
    """Mark the cells in the mesh.

    For instance if you want to mark this mesh accoring to
    the  AHA 17-segment model, then nsector = [6,6,4,1],
    i.e 6 basal, 6 mid, 4 apical and one apex

     """

    fun = dolfin.MeshFunction("size_t", mesh, 3)
    nlevels = len(nsectors)

    pi = np.pi

    assert nlevels <= 4

    if nlevels == 4:
        mus = [90, 60, 30, 10, 0]
    elif nlevels == 3:
        mus = [90, 60, 30, 0]
    elif nlevels == 2:
        mus = [90, 45, 0]
    else:
        mus = [90, 0]

    thetas = [np.linspace(pi, 3 * pi, s + 1)[:-1].tolist() + [pi] for s in nsectors]

    start = 0
    end = nsectors[0]
    regions = np.zeros((sum(nsectors), 4))
    for i in range(nlevels):
        regions.T[0][start:end] = mus[i] * pi / 180
        regions.T[3][start:end] = mus[i + 1] * pi / 180
        if i != len(nsectors) - 1:
            start += nsectors[i]
            end += nsectors[i + 1]

    start = 0
    for j, t in enumerate(thetas):
        for i in range(nsectors[j]):
            regions.T[1][i + start] = t[i]
            regions.T[2][i + start] = t[i + 1]
        start += nsectors[j]

    sfun = mark_cell_function(fun, mesh, foc, regions)

    if mark_mesh:
        # Mark the cells accordingly
        for cell in dolfin.cells(mesh):
            mesh.domains().set_marker((cell.index(), sfun[cell]), 3)

    return sfun


def mark_cell_function(fun, mesh, foc, regions):
    """
    Iterates over the mesh and stores the
    region number in a meshfunction
    """

    if foc is None:
        foc = estimate_focal_point(mesh)

    for cell in dolfin.cells(mesh):

        # Get coordinates to cell midpoint
        x = cell.midpoint().x()
        y = cell.midpoint().y()
        z = cell.midpoint().z()

        T = cartesian_to_prolate_ellipsoidal(x, y, z, foc)

        fun[cell] = strain_region_number(T, regions)

    return fun


def strain_region_number(T, regions):
    """
    For a given point in prolate coordinates,
    return the region it belongs to.

    :param regions: Array of all coordinates for the strain
                    regions taken from the strain mesh.
    :type regions: :py:class:`numpy.array`

    :param T: Some value i prolate coordinates
    :type T: :py:class:`numpy.array`

    Resturn the region number that
    T belongs to
    """

    """
    The cricumferential direction is a bit
    tricky because it goes from -pi to pi.
    To overcome this we add pi so that the
    direction goes from 0 to 2*pi
    """

    lam, mu, theta = T

    theta = theta + np.pi

    levels = get_level(regions, mu)

    if np.shape(regions)[0] + 1 in levels:
        return np.shape(regions)[0] + 1

    sector = get_sector(regions, theta)

    assert len(np.intersect1d(levels, sector)) == 1

    return np.intersect1d(levels, sector)[0] + 1


def get_level(regions, mu):

    A = np.intersect1d(
        np.where((regions.T[3] <= mu))[0], np.where((mu <= regions.T[0]))[0]
    )
    if len(A) == 0:
        return [np.shape(regions)[0] + 1]
    else:
        return A


def get_sector(regions, theta):

    if not (
        np.count_nonzero(regions.T[1] <= regions.T[2]) >= 0.5 * np.shape(regions)[0]
    ):
        raise ValueError("Surfaces are flipped")

    sectors = []
    for i, r in enumerate(regions):

        if r[1] == r[2]:
            sectors.append(i)
        else:
            if r[1] > r[2]:
                if theta > r[1] or r[2] > theta:
                    sectors.append(i)

            else:
                if r[1] < theta < r[2]:
                    sectors.append(i)

    return sectors


def estimate_focal_point(mesh):
    r"""Copmute the focal point based on approximating the
    endocardial surfaces as a ellipsoidal cap.

    .. math::

           focal = \sqrt{ l^2 - s^2}


    Arguments
    ---------
    mesh: `dolfin.mesh`
        The mesh

    Returns
    -------
    focal_point: float
        The focal point

    """

    max_coord = np.max(mesh.coordinates(), 0)
    min_coord = np.min(mesh.coordinates(), 0)

    axis = np.abs(max_coord - min_coord)
    long_axis = np.max(axis)
    short_axis = np.min(axis)

    focal = np.sqrt(long_axis ** 2 - short_axis ** 2)

    return focal



from .material_model import Material

# Material models
from .holzapfelogden import HolzapfelOgden
from .guccione import Guccione
from .linearelastic import LinearElastic
from .neohookean import NeoHookean
from .stvenantkirchhoff import StVenantKirchhoff

material_models = (
    HolzapfelOgden,
    Guccione,
    NeoHookean,
    LinearElastic,
    StVenantKirchhoff,
)

material_model_names = [m.name for m in material_models]


def get_material_model(material_model):

    for model in material_models:

        if model.name == material_model:
            return model

    raise ValueError(
        ("Material model {} does not exist. " "Please use one of {}").format(
            material_model, material_model_names
        )
    )


# Active models
from .active_strain import ActiveStrain
from .active_stress import ActiveStress



import warnings as _warnings

try:
    from ffc.quadrature.deprecation import QuadratureRepresentationDeprecationWarning

    _warnings.filterwarnings(
        "ignore", category=QuadratureRepresentationDeprecationWarning
    )
    _warnings.filterwarnings("ignore", category=DeprecationWarning)
except:
    pass

_warnings.filterwarnings("ignore", category=FutureWarning)
_warnings.filterwarnings("ignore", category=UserWarning)

from .setup_parameters import parameters
from . import setup_parameters

from .utils import annotation

setup_parameters.setup_general_parameters()

from collections import namedtuple

Patient = namedtuple("Patient", ["geometry", "data"])

from . import utils
from . import dolfin_utils
from . import io_utils
from . import numpy_mpi
from . import kinematics
from . import mechanicsproblem
from . import iterate
from . import unloader

# Subpackages
from . import material
from .material import *


from .unloader import MeshUnloader, RaghavanUnloader, FixedPointUnloader
from .geometry import (
    Geometry,
    Marker,
    CRLBasis,
    HeartGeometry,
    Microstructure,
    MarkerFunctions,
)
from .example_meshes import mesh_paths
from pulse.mechanicsproblem import (
    MechanicsProblem,
    BoundaryConditions,
    NeumannBC,
    RobinBC,
)
from .dolfin_utils import QuadratureSpace, MixedParameter, RegionalParameter

from .kinematics import (
    SecondOrderIdentity,
    DeformationGradient,
    Jacobian,
    GreenLagrangeStrain,
    LeftCauchyGreen,
    RightCauchyGreen,
    EulerAlmansiStrain,
    Invariants,
    PiolaTransform,
    InversePiolaTransform,
)
from .__version__ import __version__

__author__ = "Henrik Finsberg"
__credits__ = ["Henrik Finsberg"]
__license__ = "LGPL-3"
__maintainer__ = "Henrik Finsberg"
__email__ = "henriknf@simula.no"

__all__ = [
    "parameters",
    "setup_parameters",
    "annotation",
    "utils",
    "dolfin_utils",
    "io_utils",
    "numpy_mpi",
    "kinematics",
    "mechanicsproblem",
    "iterate",
    "unloader",
    "Geometry",
    "Marker",
    "CRLBasis",
    "HeartGeometry",
    "Microstructure",
    "MarkerFunctions",
    "QuadratureSpace",
    "MixedParameter",
    "RegionalParameter",
    "material",
    "MeshUnloader",
    "RaghavanUnloader",
    "FixedPointUnloader",
    "BoundaryConditions",
    "MechanicsProblem",
    "NeumannBC",
    "RobinBC",
    "ActiveStrain",
    "ActiveStress",
    "Material",
    "HolzapfelOgden",
    "Guccione",
    "LinearElastic",
    "NeoHookean",
    "StVenantKirchhoff",
    "SecondOrderIdentity",
    "DeformationGradient",
    "Jacobian",
    "GreenLagrangeStrain",
    "LeftCauchyGreen",
    "RightCauchyGreen",
    "EulerAlmansiStrain",
    "Invariants",
    "PiolaTransform",
    "InversePiolaTransform",
    "__version__",
]

__author__ = "Henrik Finsberg"
__credits__ = ["Henrik Finsberg"]
__license__ = "LGPL-3"
__maintainer__ = "Henrik Finsberg"
__email__ = "henriknf@simula.no"



import dolfin
from ..dolfin_utils import subplus, heaviside
from .material_model import Material


class HolzapfelOgden(Material):
    r"""
    Transversally isotropic version of the
    Holzapfel and Ogden material model

    .. math::

       \mathcal{W}(I_1, I_{4f_0})
       = \frac{a}{2 b} \left( e^{ b (I_1 - 3)}  -1 \right)
       + \frac{a_f}{2 b_f} \left( e^{ b_f (I_{4f_0} - 1)_+^2} -1 \right)

    where

    .. math::

       (\cdot)_+ = \max\{x,0\}


    .. rubric:: Reference

    [1] Holzapfel, Gerhard A., and Ray W. Ogden.
    "Constitutive modelling of passive myocardium:
    a structurally based framework for material characterization.
    "Philosophical Transactions of the Royal Society of London A:
    Mathematical, Physical and Engineering Sciences 367.1902 (2009): 3445-3475.

    """
    name = "holzapfel_ogden"

    @staticmethod
    def default_parameters():
        """
        Default matereial parameter for the Holzapfel Ogden model

        Taken from Table 1 row 3 of [1]
        """

        return {"a": 2.28, "a_f": 1.685, "b": 9.726, "b_f": 15.779}

    # def CauchyStress(self, F, p=None, deviatoric=False):

    #     I1  = self.active.I1(F)
    #     I4f = self.active.I4(F)

    #     # Active stress
    #     wactive = self.active.Wactive(F, diff = 1)

    #     dim = get_dimesion(F)
    #     I = Identity(dim)
    #     w1   = self.W_1(I1, diff = 1, dim = dim)
    #     w4f  = self.W_4(I4f, diff = 1)

    #     Fe = self.active.Fe(F)
    #     Be = Fe*Fe.T
    #     B = F*F.T

    #     fe = Fe*self.active.get_component("fiber")
    #     f = F*self.active.get_component("fiber")

    #     fefe = outer(fe, fe)
    #     ff = outer(f,f)

    #     # T = 2*w1*Be + 2*w4f*fefe + wactive*ff
    #     # T = 2*w4f*ff + wactive*ff  #2*w1*B
    #     T = wactive*ff
    #     if deviatoric:
    #         return T - tr(T)*I

    #     if p is None:
    #         return T

    #     else:
    #         return T - p*I

    def W_1(self, I_1, diff=0, *args, **kwargs):
        r"""
        Isotropic contribution.

        If `diff = 0`, return

        .. math::

           \frac{a}{2 b} \left( e^{ b (I_1 - 3)}  -1 \right)

        If `diff = 1`, return

        .. math::

           \frac{a}{b} e^{ b (I_1 - 3)}

        If `diff = 2`, return

        .. math::

           \frac{a b}{2}  e^{ b (I_1 - 3)}

        """

        a = self.a
        b = self.b

        if diff == 0:
            return a / (2.0 * b) * (dolfin.exp(b * (I_1 - 3)) - 1)
        elif diff == 1:
            return a / 2.0 * dolfin.exp(b * (I_1 - 3))
        elif diff == 2:
            return a * b / 2.0 * dolfin.exp(b * (I_1 - 3))

    def W_4(self, I_4, diff=0, *args, **kwargs):
        r"""
        Anisotropic contribution.

        If `diff = 0`, return

        .. math::

           \frac{a_f}{2 b_f} \left( e^{ b_f (I_{4f_0} - 1)_+^2} -1 \right)

        If `diff = 1`, return

        .. math::

           a_f (I_{4f_0} - 1)_+ e^{ b_f (I_{4f_0} - 1)^2}

        If `diff = 2`, return

        .. math::

           a_f h(I_{4f_0} - 1) (1 + 2b(I_{4f_0} - 1))
           e^{ b_f (I_{4f_0} - 1)_+^2}

        where

        .. math::

           h(x) = \frac{\mathrm{d}}{\mathrm{d}x} \max\{x,0\}

        is the Heaviside function.

        """
        a = self.a_f
        b = self.b_f

        if I_4 == 0:
            return 0

        if diff == 0:
            return (
                a
                / (2.0 * b)
                * heaviside(I_4 - 1)
                * (dolfin.exp(b * pow(I_4 - 1, 2)) - 1)
            )

        elif diff == 1:
            return a * subplus(I_4 - 1) * dolfin.exp(b * pow(I_4 - 1, 2))
        elif diff == 2:
            return (
                a
                * heaviside(I_4 - 1)
                * (1 + 2.0 * b * pow(I_4 - 1, 2))
                * dolfin.exp(b * pow(I_4 - 1, 2))
            )



#!/usr/bin/env python
import dolfin
import logging

parameters = dolfin.Parameters("Pulse_parameters")
parameters.add("log_level", dolfin.get_log_level())


def setup_general_parameters():
    """
    Parameters to speed up the compiler
    """

    # Parameter for the compiler
    flags = ["-O3", "-ffast-math", "-march=native"]
    dolfin.parameters["form_compiler"]["quadrature_degree"] = 4
    dolfin.parameters["form_compiler"]["representation"] = "uflacs"
    dolfin.parameters["form_compiler"]["cpp_optimize"] = True
    dolfin.parameters["form_compiler"]["cpp_optimize_flags"] = " ".join(flags)

    # dolfin.set_log_active(False)
    dolfin.set_log_level(logging.INFO)


def setup_unloading_parameters():
    """
    Parameters for coupled unloading/material parameter
    estimation.

    For info about the different parameters,
    see the unloading module.
    """

    params = dolfin.Parameters("Unloading_parameters")

    params.add("method", "fixed_point", ["fixed_point", "raghavan"])
    # Terminate if difference in reference (unloaded) volume
    # is less than tol
    params.add("tol", 0.05)
    # Maximum number of coupled iterations
    params.add("maxiter", 5)
    # Apply conitinuation step
    params.add("continuation", False)
    # Estimate initial guess based on loaded configuration
    params.add("estimate_initial_guess", True)

    unload_options = dolfin.Parameters("unload_options")
    unload_options.add("maxiter", 10)
    unload_options.add("tol", 0.01)
    unload_options.add("ub", 2.0)
    unload_options.add("lb", 0.5)
    unload_options.add("regen_fibers", False)

    params.add(unload_options)

    return params



import dolfin
from .. import kinematics
from .active_model import ActiveModel, check_component


def Wactive_transversally(Ta, C, f0, eta=0.0):
    """
    Return active strain energy when activation is only
    working along the fibers, with a possible transverse
    component defined by eta

    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A scalar function representng the mangnitude of the
        active stress in the reference configuration (firt Pioala)
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        active stress
    eta : float
        Amount of active stress in the transverse direction
        (relative to f0)
    """

    I4f = dolfin.inner(C * f0, f0)
    I1 = dolfin.tr(C)
    return dolfin.Constant(0.5) * Ta * ((I4f - 1) + eta * ((I1 - 3) - (I4f - 1)))


def Wactive_orthotropic(Ta, C, f0, s0, n0):
    """Return active strain energy for an orthotropic
    active stress

    
    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A vector function representng the mangnitude of the
        active stress in the reference configuration (firt Pioala).
        Ta = (Ta_f0, Ta_s0, Ta_n0)
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        first component
    s0 : dolfin.Function
        A vector function representng the direction of the
        second component
    n0 : dolfin.Function
        A vector function representng the direction of the
        third component
    """
    I4f = dolfin.inner(C * f0, f0)
    I4s = dolfin.inner(C * s0, s0)
    I4n = dolfin.inner(C * n0, n0)

    I4 = dolfin.as_vector([I4f - 1, I4s - 1, I4n - 1])
    return dolfin.Constant(0.5) * dolfin.inner(Ta, I4)


def Wactive_anisotropic(Ta, C, f0, s0, n0):
    """Return active strain energy for a fully anisotropic
    acitvation.
    Note that the three basis vectors are assumed to be
    orthogonal
    
    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A full tensor function representng the active stress tensor
        of the active stress in the reference configuration
        (firt Pioala).
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        first component
    s0 : dolfin.Function
        A vector function representng the direction of the
        second component
    n0 : dolfin.Function
        A vector function representng the direction of the
        third component
    """
    I4f = dolfin.inner(C * f0, f0)
    I4s = dolfin.inner(C * s0, s0)
    I4n = dolfin.inner(C * n0, n0)

    I8fs = dolfin.inner(C * f0, s0)
    I8fn = dolfin.inner(C * f0, n0)
    I8sn = dolfin.inner(C * s0, n0)

    A = dolfin.as_matrix(
        (
            (0.5 * (I4f - 1), I8fs, I8fn),
            (I8fs, 0.5 * (I4s - 1), I8sn),
            (I8fn, I8sn, 0.5 * (I4n - 1)),
        )
    )

    return dolfin.inner(Ta, A)


class ActiveStress(ActiveModel):
    """
    Active stress model
    """

    _model = "active_stress"

    def __init__(self, *args, **kwargs):

        # Fraction of transverse stress
        # (0 = active only along fiber, 1 = equal
        # amout of tension in all directions)
        self._eta = dolfin.Constant(kwargs.pop("eta", 0.0))

        self.active_isotropy = kwargs.pop("active_isotropy", "transversally")

        ActiveModel.__init__(self, *args, **kwargs)

    @property
    def eta(self):
        return self._eta

    def Wactive(self, F, diff=0):
        """Active stress energy
        """

        C = F.T * F

        if diff == 0:

            if self.active_isotropy == "transversally":
                return Wactive_transversally(
                    Ta=self.activation_field, C=C, f0=self.f0, eta=self.eta
                )

            elif self.active_isotropy == "orthotropic":
                return Wactive_orthotropic(
                    Ta=self.activation_field, C=C, f0=self.f0, s0=self.s0, n0=self.n0
                )

            elif self.active_isotropy == "fully_anisotropic":

                return Wactive_anisotropic(
                    Ta=self.activation_field, C=C, f0=self.f0, s0=self.s0, n0=self.n0
                )
            else:
                msg = ("Unknown acitve isotropy " "{}").format(self.active_isotropy)
                raise ValueError(msg)

        elif diff == 1:
            return self.activation_field

    @property
    def type(self):
        return "ActiveStress"

    def I1(self, F, *args):
        return self._I1(F)

    def I4(self, F, component="f0", *args):

        check_component(component)
        a0 = getattr(self, component)
        return self._I4(F, a0)

    @property
    def Fa(self):
        return kinematics.SecondOrderIdentity(self.f0)

    def Fe(self, F):
        return F



#!/usr/bin/env python

import os
import logging

import dolfin
from . import parameters

if dolfin.__version__.startswith("20"):
    # Year based versioning
    DOLFIN_VERSION_MAJOR = float(dolfin.__version__.split(".")[0])
else:
    DOLFIN_VERSION_MAJOR = float(".".join(dolfin.__version__.split(".")[:2]))


class Annotation(object):
    """
    Object holding global annotation for dolfin-adjoint
    """

    def __init__(self):
        self.annotate = False

    @property
    def annotate(self):
        return self._annotate

    @annotate.setter
    def annotate(self, annotate=False):
        """
        Set global annotation for dolfin-adjoint.
        Default False
        """
        if "adjoint" in dolfin.parameters:
            dolfin.parameters["adjoint"]["stop_annotating"] = not annotate
        else:
            try:
                from pyadjoint.tape import _stop_annotating
            except ImportError:
                # Dolfin-adjoint is most likely not installed
                pass
            else:
                _stop_annotating = not annotate

        # Update local variable
        self._annotate = annotate


annotation = Annotation()


def mpi_comm_world():
    if DOLFIN_VERSION_MAJOR >= 2018:
        return dolfin.MPI.comm_world
    else:
        return dolfin.mpi_comm_world()


def value_size(obj):
    try:
        return obj.value_size()
    except AttributeError:
        value_shape = obj.value_shape()
        if len(value_shape) == 0:
            return 1
        else:
            return [0]


def set_default_none(NamedTuple, default=None):
    NamedTuple.__new__.__defaults__ = (default,) * len(NamedTuple._fields)


# Dummy object
class Object(object):
    pass


def make_logger(name, level=parameters["log_level"]):
    def log_if_process0(record):
        if dolfin.MPI.rank(mpi_comm_world()) == 0:
            return 1
        else:
            return 0

    mpi_filt = Object()
    mpi_filt.filter = log_if_process0

    logger = logging.getLogger(name)
    logger.setLevel(level)

    ch = logging.StreamHandler()
    ch.setLevel(0)
    # formatter = logging.Formatter('%(message)s')
    formatter = logging.Formatter(
        ("%(asctime)s - " "%(name)s - " "%(levelname)s - " "%(message)s")
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.addFilter(mpi_filt)

    dolfin.set_log_level(logging.WARNING)

    ffc_logger = logging.getLogger("FFC")
    ffc_logger.setLevel(logging.WARNING)
    ffc_logger.addFilter(mpi_filt)

    ufl_logger = logging.getLogger("UFL")
    ufl_logger.setLevel(logging.WARNING)
    ufl_logger.addFilter(mpi_filt)

    return logger


logger = make_logger(__name__)


def number_of_passive_controls(params):
    # Number of passive parameters to optimize
    return sum([not v for v in params["Fixed_parameters"].values()])


def get_lv_marker(geometry):

    if "ENDO" in geometry.markers:
        return geometry.markers["ENDO"][0]
    elif "ENDO_LV" in geometry.markers:
        return geometry.markers["ENDO_LV"][0]
    else:
        raise KeyError("Geometry is missing marker for LV ENDO")


def save_logger(params):

    outdir = os.path.dirname(params["sim_file"])
    logfile = "output.log" if outdir == "" else outdir + "/output.log"
    logging.basicConfig(
        filename=logfile,
        filemode="a",
        format="%(message)s",
        datefmt="%H:%M:%S",
        level=logging.INFO,
    )

    ffc_logger = logging.getLogger("FFC")
    ffc_logger.setLevel(logging.WARNING)
    ufl_logger = logging.getLogger("UFL")
    ufl_logger.setLevel(logging.WARNING)

    import datetime

    time = datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d %H:%M:%S")
    logger.info("Time: {}".format(time))


class UnableToChangePressureExeption(Exception):
    pass


class AutoVivification(dict):
    """Implementation of perl's autovivification feature."""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value


def print_head(for_res, display_iter=True):

    targets = for_res["optimization_targets"]
    keys = targets.keys() + ["regularization"]
    n = len(keys)

    head = "\n{:<6}\t".format("Iter") if display_iter else "\n" + " " * 7
    head += (
        "{:<10}\t".format("Obj")
        + "{:<10}".format("||grad||")
        + "\t"
        + (n * "I_{:<10}\t").format(*keys)
    )

    return head


def print_line(for_res, it=None, grad_norm=None, func_value=None):

    func_value = for_res["func_value"] if func_value is None else func_value
    grad_norm = 0.0 if grad_norm is None else grad_norm

    targets = for_res["target_values"]
    reg_func = targets.pop("regularization")
    values = targets.values() + [reg_func]
    targets["regularization"] = reg_func

    n = len(values)
    line = "{:<6d}\t".format(it) if it is not None else ""
    line += (
        "{:<10.2e}\t".format(func_value)
        + "{:<10.2e}".format(grad_norm)
        + "\t"
        + (n * "{:<10.2e}\t").format(*values)
    )

    return line


def rename_attribute(object_, old_attribute_name, new_attribute_name):
    setattr(object_, new_attribute_name, getattr(object_, old_attribute_name))
    delattr(object_, old_attribute_name)


class TablePrint(object):
    """
    Print output in nice table format.

    **Example of use**::

      fldmap = (
         'LVP',  '0.5f',
         'LV_Volume', '0.5f',
         'Target_Volume', '0.5f',
         'I_strain', '0.2e',
         'I_volume', '0.2e',
         'I_reg', '0.2e',
         )

      my_print = TablePrint(fldmap)
      print my_print.print_head()
      print my_print.print_line(LVP=1, LV_Volume=1, Target_Volume=1,
                                I_strain=1, I_volume=1, I_reg=1)

    """

    def __init__(self, fldmap, fancyhead=False):

        if fancyhead:
            q = [int(a.split(".")[0]) for a in fldmap[1::2]]

            fmt = "\t".join(["{:" + "{}".format(f) + "}" for f in q])

            self.head = fmt.format(*fldmap[0::2])
        else:
            self.head = "\n" + "\t".join(fldmap[0 : len(fldmap) : 2])

        self.fmt = "\t".join(
            [
                "{" + "{0}:{1}".format(col, f) + "}"
                for col, f in zip(
                    fldmap[0 : len(fldmap) : 2], fldmap[1 : len(fldmap) : 2]
                )
            ]
        )

    def print_head(self):
        return self.head

    def print_line(self, **kwargs):
        return self.fmt.format(**kwargs)


class Text:
    """
    Ansi escape sequences for coloured text output
    """

    _PURPLE = "\033[95m"
    _OKBLUE = "\033[94m"
    _OKGREEN = "\033[92m"
    _YELLOW = "\033[93m"
    _RED = "\033[91m "
    _ENDC = "\033[0m"

    @staticmethod
    def blue(text):
        out = Text._OKBLUE + text + Text._ENDC
        return out

    @staticmethod
    def green(text):
        out = Text._OKGREEN + text + Text._ENDC
        return out

    @staticmethod
    def red(text):
        out = Text._RED + text + Text._ENDC
        return out

    @staticmethod
    def yellow(text):
        out = Text._YELLOW + text + Text._ENDC
        return out

    @staticmethod
    def purple(text):
        out = Text._PURPLE + text + Text._ENDC
        return out

    @staticmethod
    def decolour(text):
        to_remove = [
            Text._ENDC,
            Text._OKBLUE,
            Text._OKGREEN,
            Text._RED,
            Text._YELLOW,
            Text._PURPLE,
        ]

        for chars in to_remove:
            text = text.replace(chars, "")
        return text



#!/usr/bin/env python
# c) 2001-2017 Simula Research Laboratory ALL RIGHTS RESERVED
# Authors: Henrik Finsberg
# END-USER LICENSE AGREEMENT
# PLEASE READ THIS DOCUMENT CAREFULLY. By installing or using this
# software you agree with the terms and conditions of this license
# agreement. If you do not accept the terms of this license agreement
# you may not install or use this software.

# Permission to use, copy, modify and distribute any part of this
# software for non-profit educational and research purposes, without
# fee, and without a written agreement is hereby granted, provided
# that the above copyright notice, and this license agreement in its
# entirety appear in all copies. Those desiring to use this software
# for commercial purposes should contact Simula Research Laboratory AS:
# post@simula.no
#
# IN NO EVENT SHALL SIMULA RESEARCH LABORATORY BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE
# "PULSE-ADJOINT" EVEN IF SIMULA RESEARCH LABORATORY HAS BEEN ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE. THE SOFTWARE PROVIDED HEREIN IS
# ON AN "AS IS" BASIS, AND SIMULA RESEARCH LABORATORY HAS NO OBLIGATION
# TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# SIMULA RESEARCH LABORATORY MAKES NO REPRESENTATIONS AND EXTENDS NO
# WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESSED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
import dolfin

try:
    from dolfin_adjoint import Constant
except ImportError:
    from dolfin import Constant

from .. import kinematics
from ..dolfin_utils import RegionalParameter


def check_component(component):
    components = ("f0", "s0", "n0")
    msg = ("Component must be one " "of {}, got {}").format(component, components)
    assert component in components, msg


class ActiveModel(kinematics.Invariants):
    def __init__(
        self,
        activation=None,
        f0=None,
        s0=None,
        n0=None,
        T_ref=None,
        isochoric=True,
        *args,
        **kwargs
    ):

        # Fiber system
        self.f0 = f0
        self.s0 = s0
        self.n0 = n0

        self._activation = (
            Constant(0, name="activation") if activation is None else activation
        )

        self.T_ref = (
            Constant(T_ref, name="T_ref") if T_ref else Constant(1.0, name="T_ref")
        )

        kinematics.Invariants.__init__(self, isochoric, *args)

    @property
    def model_type(self):
        return self._model

    def Wactive(self, *args, **kwargs):
        return 0

    @property
    def eta(self):
        return 0

    @property
    def activation_field(self):
        """
        Return the activation field.
        If regional, this will return a piecewise
        constant function (DG_0)
        """

        # Activation
        if isinstance(self._activation, RegionalParameter):
            # This means a regional activation
            # Could probably make this a bit more clean
            activation = self._activation.function
        else:
            activation = self._activation

        return self.T_ref * activation

    @property
    def activation(self):
        """
        Return the activation paramter.
        If regional, this will return one parameter
        for each segment.
        """
        return self._activation

    @activation.setter
    def activation(self, f):
        self._activation = f

    def I1(self, F):
        return self._I1(F)

    def I2(self, F):
        return self._I2(F)

    def I3(self, F):
        return self._I3(F)

    def I4(self, F, a0):
        return self._I4(F, a0)

    def I5(self, F, a0):
        return self._I5(F, a0)

    def I6(self, F, a0):
        return self._I6(F, a0)

    def I7(self, F, a0):
        return self._I7(F, a0)

    def I8(self, F, a0, b0):
        return self._I8(F, a0, b0)
